/*
*  topf2mvix
*  
*  Make Topfield .rec files playable on Unicorn MVIX 760HD media player.
*  File extension must be .ts to be recognized by MVIX (as of fw 1.1.27).
*
*  Adapted by RAGtime for TF4000/PATPMT header option!
*
*  June 2007, hifan@gmx.de
*/

#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>

// constants
const int TS_SIZE = 188;
const int NUM_ANALYZE = 3000;
const int PID_TABLE_SIZE = 10;
const int OFFSET_VIDEO = 18; // 0-based
const int OFFSET_AUDIO = 23; // 0-based
/* total header length (TF4000: 3pakets=564bytes & TF5000: 20pakets=3760bytes) */
const int HEADER_TF4000 = 3;
const int HEADER_TF5000 = 20;

// data (sure that this is correct? RAG)
unsigned char PAT[] = {0x47,0x40,0x00,0x11,0x00,0x00,0xb0,0x0d,0x00,0x01,0xc3,0x00,0x00,0x00,0x01,0xe1,0x00,0x76,0x57,0x8e,0x5f};
int PAT_LEN = 21;
unsigned char PMT[] = {0x47,0x41,0x00,0x11,0x00,0x02,0xb0,0x17,0x00,0x01,0xc1,0x00,0x00,0xe0,0xc9,0xf0,0x00,0x02,0xe0,0x33,0xf0,0x00,0x03,0xe0,0x44,0xf0,0x00,0x81,0x3b,0x1b,0xa4};
int PMT_LEN = 31;
//unsigned char NUL[] = {0x47,0x1f,0xff,0x11};
//int NUL_LEN = 4;
// jkIT means should work without the 0x11 
// (Payload-Flag set and Continuity-Counter=1 makes no sense for a NULL-Paket with PID=0x1fff,
// tootired said we could additionally set the transport error indicator bit ==> PID=0x9fff)
unsigned char NUL[] = {0x47,0x1f,0xff};
int NUL_LEN = 3;
// and tootired said also this could work (without any PID)?
//unsigned char NUL[] = {0x47};
//int NUL_LEN = 1;

// set header length later according to CLI (default is TF5000 headers)
int HEADER_LEN = 0;

// default is don't write any PAT/PMT at beginning of header 
int PATPMT_LEN = 0;

// default is don't rewrite PAT/PMT throughout the file
int REWRITE = 0;

// globals
int video;
int audio;
int pidtable[99];
int pidcnt[99];
int pmtpid = -1;
     
// main
int main(int argc,  char *argv[])
{
  char filename[250];
  int audiopid = 0, i;

  if (argc == 1)
  {
    fprintf(stderr, "Patch Header of Topfield Transport Streams:\n");
    fprintf(stderr, " topf2ts <filename> [<audiopid>] ['REWRITE'] ['TF4000'] ['PATPMT']\n");
    exit(1);
  } 

  // the filename has to be the FIRST argument!!!
  strncpy(filename, argv[1], 250);
  
  /* default is TF5000 */
  HEADER_LEN = HEADER_TF5000;
  for(i=2;i<argc;i++) {
    if(strstr(argv[i], "TF4000") || strstr(argv[i], "tf4000")) 
      HEADER_LEN = HEADER_TF4000;
    else if(strstr(argv[i], "PATPMT") || strstr(argv[i], "patpmt"))
      PATPMT_LEN = 2;
    else if(strstr(argv[i], "REWRITE")|| strstr(argv[i], "rewrite"))
      REWRITE = 1;
    else // no other option? expect an hex audio PID...
      audiopid = strtol(argv[i], NULL, 16);   
  }

  // 1 - analyze file data
  analyze(filename, audiopid);
  
  // 2 - re-write header
  writeHeader(filename);
  
  // 3 - overwrite PAT/PMT throughout file?
  if(REWRITE)
    rewritePatPmt(filename);
  
  // rename if necessary (i.e. not already ending with .ts?)
  doRename(filename);
  
  printf("end.\n");
}

analyze(char *filename, int audiopid)
{
  FILE *fp;
  int i, c, cnt, pid, c1, cnt2, sync=0, corrupted=0;
  int patOffset = -1;
  
  for(i = 0; i<PID_TABLE_SIZE; i++)
  {
    pidtable[i] = 0;
    pidcnt[i] = 0;
  }

  fp = fopen64(filename, "r"); 

  if (fp == NULL)
  {
    fprintf(stderr,"Cannot open %s\n", filename);
    exit(1);
  }

  // go to offset HEADER_LEN * TS_SIZE - beginning of TS data
  fseeko64(fp, HEADER_LEN * TS_SIZE, SEEK_SET);
  fprintf(stderr,"Skipped %d pakets of topfield header...\n",HEADER_LEN);
  
  for(cnt = 0; cnt < NUM_ANALYZE; cnt++) // inspect first x packets
  {
    // corrupted paket? seek the next 188 bytes for a 0x47 to get in sync!
    for(cnt2 = 0; cnt2 < TS_SIZE; cnt2++) 
    {
      c = getc(fp); // skip 0x47 marker
      if (c == EOF) 
        goto BREAK_ALL_LOOPS;    
      if (c == 0x47) 
        break;
    }
    
    if (cnt2 >= TS_SIZE) 
    {  
      fprintf(stderr,"Invalid file! (no 0x47 @ paket %d found!)\n",cnt);
      exit(1);
    }   
    else if(cnt2>0) {
      fprintf(stderr,"Corrupted paket? (Found begin of %d. TS-paket @ position %d)\n",HEADER_LEN+1+cnt,cnt2);
      sync=0;
      corrupted++;
      // hmmmm,... is PID OK, if this paket is also to short?	
    } 
    else if(!sync)
    { 
      fprintf(stderr,"We seem to be in sync now...\n");     
      sync=1;
    }  
    
    c = fgetc(fp);
    c1 = fgetc(fp);
    pid = getPid(c, c1);
    if (pid == 0 && patOffset < 0) // remember a PAT
      patOffset = cnt;
      // printf("found PID: %x\n", pid);
    
    addPid(pid, pidtable, pidcnt);

    fseeko64(fp, TS_SIZE - 3, SEEK_CUR); // read on to next TS packet
  }
  // the only valid usage of a goto: breaking nested loops! ;-)
  BREAK_ALL_LOOPS:
  fprintf(stderr,"%d pakets analyzed, %d corrupted found!\n",NUM_ANALYZE,corrupted);

  if (patOffset >= 0)
  {
    pmtpid = findPmtPid(fp, patOffset);
    //printf("pmt-pid: %4x\n", pmtpid);
  }

  evalPids(pidtable, pidcnt, audiopid);

  fclose(fp);
}

addPid(int pid, int pidtable[], int pidcnt[])
{
  int i;

  if (pid == 0)
    return; // ignore PAT packets
    
  for(i = 0; i<PID_TABLE_SIZE; i++)
  {
    if (pidtable[i] == pid)
    {
      pidcnt[i]++;
      break;
    }
    if (pidtable[i] == 0)
    {
      pidtable[i] = pid;
      pidcnt[i]++;
      break;
    }
  }
}

evalPids(int pidtable[], int pidcnt[], int audiopid)
{
  int first = 0, second = 0, firstcnt = 0, secondcnt = 0;
  int i;
  for(i = 0; i<PID_TABLE_SIZE; i++)
  {
    // printf("%4x: %i\n", pidtable[i], pidcnt[i]);
    if (pidcnt[i] > firstcnt)
    {
      secondcnt = firstcnt;
      second = first;
      firstcnt = pidcnt[i];
      first = pidtable[i];
    }
    else if (pidcnt[i] > secondcnt)
    {
      secondcnt = pidcnt[i];
      second = pidtable[i];
    }
  }  
  
  // printf("first: %4x (%i)\n", first, firstcnt);
  // printf("secnd: %4x (%i)\n", second, secondcnt);
  
  video = first;
  audio = second;
  for(i = 0; i<PID_TABLE_SIZE; i++) // prefer +1 if existent and not too small and not pmt-pid
  {
    if ((pidtable[i] == (video + 1)) && (secondcnt / pidcnt[i] < 2) 
        && ((video + 1) != pmtpid))
    {
      audio = video + 1;
      break;
    }
  }
  
  if (audiopid > 0) // manual override
    audio = audiopid;

  printf("video: %4x\n", video);
  printf("audio: %4x\n", audio);
}

writeHeader(char* filename)
{
  int i;
  
  if(PATPMT_LEN) printf("writing header PAT/PMT/%dxNULL ...\n",HEADER_LEN-PATPMT_LEN);
  else printf("writing header %dxNULL ...\n",HEADER_LEN);

  // write
  FILE *fp;
  fp = fopen64(filename, "r+b");
  if (fp == NULL)
  {
    fprintf(stderr,"Cannot open target file: %s\n", filename);
    exit(1);
  }
  
  fseeko64(fp, 0, SEEK_SET);

  if(PATPMT_LEN) {
    // set pids to PMT
    PMT[OFFSET_VIDEO] = 0xe0 | (video / 256);
    PMT[OFFSET_VIDEO + 1] = video%256;
    PMT[OFFSET_AUDIO] = 0xe0 | (audio / 256);
    PMT[OFFSET_AUDIO + 1] = audio%256;

    writePacket(fp, PAT, PAT_LEN);
    writePacket(fp, PMT, PMT_LEN);
  }

  /* How long is the TF4000PVR Header? ... 564bytes (according to jkIT)! */
  for(i = 0; i < HEADER_LEN-PATPMT_LEN; i++)
  {
    writePacket(fp, NUL, NUL_LEN);
  }
  

  fclose(fp);
}

rewritePatPmt(char* filename)
{
  FILE *fp;
  int offset = 0, c, c1, pid, pmt = 0;
  int patcnt = 0, pmtcnt = 0;

  printf("rewriting PAT/PMT in complete file..\n");

  fp = fopen64(filename, "r+b");
  if (fp == NULL)
  {
    fprintf(stderr,"Cannot open target file: %s\n", filename);
    exit(1);
  }

  while(1)
  {
    //if (fseeko64(fp, (20 + offset) * TS_SIZE, SEEK_SET) != 0)
    if (fseeko64(fp, (HEADER_LEN + offset) * TS_SIZE, SEEK_SET) != 0)
      break;

    c = fgetc(fp);
    if (c == EOF) break;

    if (c != 0x47)
    {
      fprintf(stderr,"Error in file (no TS) at packet offset: %i\n", offset);
      break;
    }

    c = fgetc(fp);
    c1 = fgetc(fp);
    pid = getPid(c, c1);

    if (pid == 0) // PAT
    {
      if (pmt == 0) // find PMT-PID
      {
        if (pmtpid > 0)
        {
          pmt = pmtpid;
        }
        else
        {
          pmt = findPmtPid(fp, offset);
          printf("pmt-pid: %4x\n", pmt);
        }
      }

      //fseeko64(fp, (20 + offset) * TS_SIZE, SEEK_SET);
      fseeko64(fp, (HEADER_LEN + offset) * TS_SIZE, SEEK_SET);
      writePacket(fp, PAT, PAT_LEN);
      patcnt++;
    }
      
    if (pid == pmt) // PMT
    {
      //fseeko64(fp, (20 + offset) * TS_SIZE, SEEK_SET);
      fseeko64(fp, (HEADER_LEN + offset) * TS_SIZE, SEEK_SET);
      writePacket(fp, PMT, PMT_LEN);
      pmtcnt++;
    }
   
    offset++; 
    if (offset%10000 == 0) 
    { 
      printf(".");
      fflush(stdout);
    }
  }
  
  printf("\nnumber of pat/pmt written: %i / %i\n", patcnt, pmtcnt);

  fclose(fp);  
}

int findPmtPid(FILE *fp, int offset)
{
  int i, j, c, c1, pmt, len;

  //fseeko64(fp, (20 + offset) * TS_SIZE + 7, SEEK_SET);
  fseeko64(fp, (HEADER_LEN + offset) * TS_SIZE + 7, SEEK_SET);
  len = fgetc(fp);
  len -= 9; // 5 more + 4 byte crc32
  len /= 4;
  printf("number of records in pat: %i\n", len);
  
  //fseeko64(fp, (20 + offset) * TS_SIZE + 13, SEEK_SET);
  fseeko64(fp, (HEADER_LEN + offset) * TS_SIZE + 13, SEEK_SET);
  for(i = 0; i < len; i++)
  {
    c = fgetc(fp);
    c1 = fgetc(fp);
    c = fgetc(fp);
    c1 = fgetc(fp);
    pmt = getPid(c, c1);
    if (pmt != 0)
    {
      for(j = 0; j < PID_TABLE_SIZE; j++)
      {
        if (pidtable[j] == pmt)
          return pmt; // found one (hopefully "it")
      }
    }
  }
  
  return -1; // nothing found
}

int getPid(int b1, int b2)
{
  int pid;
  pid = 256 * (b1 & 0x1f);
  pid += b2;
  return pid;
}
 
writePacket(FILE *fp, unsigned char *data, int len)
{
  int i;
  for (i = 0; i < len; i++)
  {
    fputc(data[i], fp);
  }
  for (i = 0; i < (TS_SIZE - len); i++)
  {
    fputc(0xff, fp);
  }
}

doRename(char* filename)
{
  char filename1[strlen(filename) + 5];
  if (!strstr(filename, ".ts"))
  {
    strcpy(filename1, filename);
    strcat(filename1, ".ts");
    rename(filename, filename1);
    printf("renamed file to: %s\n", filename1);
  }
}
